/*
 * Copyright 2016 Software Freedom Conservancy Inc.
 * Copyright 2017 Michael Gratton <mike@vee.net>
 */

body {
  margin: 0 !important;
  border: 0 !important;
  padding: 0 !important;
  color: black;
  background-color: #f9f9f9;
  font-size: medium !important;
}

body.plain, body.plain * {
  font-family: monospace !important;
  font-weight: normal;
  font-style: normal;
  font-size: medium !important;
  text-decoration: none;
}

body.plain a {
  cursor: text;
}

body > *.geary-no-display {
  display: none !important;
}

body > div#geary-body {
  margin: 0 !important;
  border: 0 !important;
  padding: 12px !important;
  outline: 0px !important;
}

body > div#geary-signature,
body > div#geary-quote {
  margin: 0 6px !important;
  border: 0 !important;
  padding: 6px !important;
}

body > div.geary-focus {
  background-color: white;
}

body > div#geary-signature.geary-focus,
body > div#geary-quote.geary-focus {
  outline: 1px dashed #ccc !important;
}

blockquote {
  margin: 0 10px;
  border: 0;
  border-left: 3px #aaa solid;
  padding: 0 5px;
}

pre {
  white-space: pre-wrap;
  margin: 0;
}
