/* Copyright 2016 Software Freedom Conservancy Inc.
 *
 * This software is licensed under the GNU Lesser General Public License
 * (version 2.1 or later).  See the COPYING file in this distribution.
 */

namespace Geary.Smtp.DataFormat {

public const string ARG_SEPARATOR = " ";

public const string LINE_TERMINATOR = "\r\n";
public const DataStreamNewlineType LINE_TERMINATOR_TYPE = DataStreamNewlineType.CR_LF;

public const string DATA_TERMINATOR = "\r\n.\r\n";

public const string CANCEL_AUTHENTICATION = "*";

}

